@@ -7,7 +7,14 @@
 import os
 import re
 from . import structs
-from .file import FILES, MySettingFile, MySetting2File, DjmMySettingFile, DevSettingFile
+from .file import (
+    FILES,
+    SettingsFile,
+    MySettingFile,
+    MySetting2File,
+    DjmMySettingFile,
+    DevSettingFile,
+)
 
 RE_MYSETTING = re.compile(".*SETTING[0-9]?.DAT$")
 
@@ -23,6 +30,6 @@ def get_mysetting_paths(root, deep=False):
     return files
 
 
-def read_mysetting_file(path):
+def read_mysetting_file(path) -> SettingsFile:
     obj = FILES[os.path.split(path)[1]]
     return obj.parse_file(path)
