@@ -15,6 +15,7 @@ import sys
 import json
 import logging
 import warnings
+import psutil
 import xml.etree.cElementTree as xml
 
 logger = logging.getLogger(__name__)
@@ -94,6 +95,98 @@ def read_pyrekordbox_configuration():
     return dict()
 
 
+def get_process_id(name: str, raise_exec=False) -> int:
+    """Returns the ID of a process if it exists.
+
+    Parameters
+    ----------
+    name : str
+        The name of the process, for example 'rekordbox'.
+    raise_exec : bool, optional
+        Raise an exception if the process can not be found.
+
+    Returns
+    -------
+    pid : int
+        The ID of the process if it exists, otherwise zero.
+
+    Raises
+    ------
+    RuntimeError: If ``raise_exec=True``, raises a runtime error if the application
+        is not running.
+
+    Examples
+    --------
+    >>> get_process_id("rekordbox")
+    12345
+
+    >>> get_process_id("rekordboxAgent")
+    23456
+    """
+    for proc in psutil.process_iter():
+        proc_name = os.path.splitext(proc.name())[0]  # needed on Windows (.exe)
+        try:
+            if proc_name == name:
+                return proc.pid
+        except psutil.AccessDenied:
+            pass
+    if raise_exec:
+        raise RuntimeError("No process with name 'rekordbox' found!")
+    return 0
+
+
+def get_rekordbox_pid(raise_exec=False):
+    """Returns the process ID of the Rekordbox application.
+
+    Parameters
+    ----------
+    raise_exec : bool, optional
+        Raise an exception if the Rekordbox process can not be found.
+
+    Returns
+    -------
+    pid : int
+        The ID of the Rekordbox process if it exists, otherwise zero.
+
+    Raises
+    ------
+    RuntimeError: If ``raise_exec=True``, raises a runtime error if the Rekordbox
+        application is not running.
+
+    Examples
+    --------
+    >>> get_rekordbox_pid()
+    12345
+    """
+    return get_process_id("rekordbox", raise_exec)
+
+
+def get_rekordbox_agent_pid(raise_exec=False):
+    """Returns the process ID of the RekordboxAgent application.
+
+    Parameters
+    ----------
+    raise_exec : bool, optional
+        Raise an exception if the RekordboxAgent process can not be found.
+
+    Returns
+    -------
+    pid : int
+        The ID of the RekordboxAgent process if it exists, otherwise zero.
+
+    Raises
+    ------
+    RuntimeError: If ``raise_exec=True``, raises a runtime error if the RekordboxAgent
+        application is not running.
+
+    Examples
+    --------
+    >>> get_rekordbox_agent_pid()
+    23456
+    """
+    return get_process_id("rekordboxAgent", raise_exec)
+
+
 def get_pioneer_app_dir(path=""):
     """Returns the path of the Pioneer application data directory.
 
@@ -283,7 +376,8 @@ def read_rekordbox6_asar(rb6_install_dir):
         location = os.path.join(root, "rekordboxAgent.app", "Contents", "Resources")
         encoding = "cp437"
     else:
-        raise logger.warning(f"OS {sys.platform} not supported!")
+        logger.warning(f"OS {sys.platform} not supported!")
+        return ""
 
     # Read asar file
     path = os.path.abspath(os.path.join(location, "app.asar"))
